open Printf
open Print
open Types


(* ------------------------------------------------------------------------- *)

(* Error messages. *)

(* For simplifity, we consider a type mismatch as a fatal error.
   Although the expected and inferred types do not match, we could still
   continue with the expected type. This sometimes allows us to report
   multiple error messages in one run. *)

let mismatch p loc expected inferred =
  Error.error [loc] (sprintf
    "Type mismatch.\nExpected: %s\nInferred: %s"
    (print_type p expected)
    (print_type p inferred)
  )

let mismatch_instance p loc expected inferred =
  Error.error [loc] (sprintf
    "Invalid existential package construction.\n\
     The inferred type is not an instance of the desired existential type.\n\
     Desired: %s\nInferred: %s"
    (print_type p expected)
    (print_type p inferred)
  )

let escapes p loc a ty =
  Error.error [loc] (sprintf
    "The type variable %s escapes its scope: it appears free in\n%s"
    (print_type p (TyFreeVar a))
    (print_type p ty)
  )

let expected_form p loc form ty =
  Error.error [loc] (sprintf
    "Type mismatch: I expected %s type.\nInferred: %s"
    form
    (print_type p ty)
  )

let expected_field p loc label ty =
  Error.error [loc] (sprintf
    "Type mismatch: there is no record field named \"%s\" in this record.\n\
     Inferred: %s"
    (print_label label)
    (print_type p ty)
  )

let expected_useless_bound p loc codomain =
  Error.error [loc] (sprintf
    "Invalid coercion\n:\n\
     The codomain of the coercion should have a useless bound.\n\
     Codomain: %s\n"
    (print_type p codomain)
  )

(* Errors for elaboration *)

let escapes_distrib p loc a ty =
  Error.error [loc] (sprintf
    "Invalid use of coercion distrib.\n\
     The type variable %s would appears free in its domain\n%s"
    (print_type p (TyFreeVar a))
    (print_type p ty)
  )

let mismatch_coercion p loc cty ty =
  Error.error [loc] (sprintf
    "Invalid coercion elimination.\n\
     Type %s\nis not an instance of type\n%s\n"
    (print_type p ty)
    (print_type p (TyQuantified (QForall, cty)))
  )

let unexpected_coercion_form p loc q mode =
  let form = 
    match q with
    | QForall -> "universal elimination form -forall"
    | QExists -> "existential introduction form +exists" in
  let dir, polarity = if mode then "<", "positive" else ">", "negative" in
  Error.error [loc] (sprintf
    "Ill formed coercion: the %s%s\n\
    should appear at a %s occurrence.\n"
    form dir polarity
  )

let unexpected_field p loc label ty =
  Error.error [loc] (sprintf
    "Type mismatch: there should be a field named \"%s\" in this record.\n\
     Inferred: %s"
    (print_label label)
    (print_type p ty)
  )
